#ifndef MPO_HTTPD_H
#define MPO_HTTPD_H

#ifdef WIN32
#pragma warning (disable:4251)	// disable the warning about std::map needing a DLL interface (since it's private, I don't think this warning applies)
#endif

#include "mpo_deleter.h"
#include "mpo_stream.h"
#include "mpo_numstr.h"	// for 64-bit number
#include "mpo_server.h"
#include "mpo_thread.h"

#include <map>
using namespace std;

typedef map<string,string> MapNameValue;

class mpo_httpd_headers
{
public:
	mpo_httpd_headers(unsigned int uStatusCode, const string &strReasonPhrase) :
	  u64ContentLength(0),
		  m_uStatusCode(uStatusCode),
		  m_strReasonPhrase(strReasonPhrase)
	  {
	  }

	  bool operator== (const mpo_httpd_headers &h) const
	  {
		  return (h.m_uStatusCode == m_uStatusCode &&
			  h.u64ContentLength == u64ContentLength &&
			  h.m_strReasonPhrase == m_strReasonPhrase &&
			  h.m_strContentType == m_strContentType &&
			  h.m_mapOtherHeaders == m_mapOtherHeaders);
	  }

	/////

	MPO_UINT64 u64ContentLength;
	unsigned int m_uStatusCode;
	string m_strReasonPhrase;
	string m_strContentType;
	MapNameValue m_mapOtherHeaders;
};

// the purpose of this interface is to make unit testing/mocking easier
class Impo_httpd
{
public:
	// MSG_OK: method has been received, so it's ok to call ParseRequest
	// MSG_TIMEOUT: method has not been received fully (so call this function again until you get MSG_OK or MSG_ERROR)
	virtual StreamMsg ReadRequest(unsigned int uTimeoutMs) = 0;

	// parses the request read by ReadRequest.
	// Returns true on success, false on failure.
	// This function must be called after ReadRequest returns MSG_OK and before any other function is called.
	virtual bool ParseRequest() = 0;

	// returns for example "GET"
	virtual string GetMethod() const = 0;

	// returns for example "/whatever?query#fragment";
	// (this string will still be escaped!)
	virtual string GetURI() const = 0;

	// returns an unescaped string
	virtual string GetPath() const = 0;	// "/whatever" from "/whatever?query#fragment"
	virtual string GetQuery() const = 0;	// "query" from "/whatever?query#fragment"
	virtual string GetFragment() const = 0;	// "fragment" from "/whatever?query#fragment"

	// returns for example "HTTP/1.1"
	virtual string GetHTTPVersion() const = 0;

	// Gets client headers such as "Host"
	// Returns false if no such header was sent by the client.
	virtual bool GetHeader(string &strValue, const string &strName) = 0;

	// Reads in POST data.
	// Returns MSG_END if no more POST data is available (or no POST data was ever available).
	virtual StreamMsg Recv(void *buffer, size_t stBytesCanRead, size_t *pstBytesRead, unsigned int uTimeoutMs) = 0;

	// sets headers in preparation to call SendHeaders
	virtual bool SetHeaders(const mpo_httpd_headers *pHeaders) = 0;

	// This function sends headers to the client after GetMethod has been called.
	// Returns MSG_OK if all headers have been sent and it's time to call Send.
	// Returns MSG_TIMEOUT if not all headers have been sent (keep calling to send them).
	virtual StreamMsg SendHeaders(unsigned int uTimeoutMs) = 0;

	virtual StreamMsg Send(const void *buffer, size_t bytes_can_send, size_t *bytes_sent, unsigned int timeout_ms) = 0;

	// returns a string describing the last error that occurred (to help troubleshoot)
	virtual string GetLastErrorStr() = 0;
};

class mpo_httpd;

typedef boost::shared_ptr<mpo_httpd> MpoHttpdSPtr;

class EXPORT_ME mpo_httpd : public Impo_httpd, public MpoDeleter
{
public:
	static MpoHttpdSPtr GetInstance(INonblockingStream *pStream);

	// MSG_OK: method has been received, so it's ok to call ParseRequest
	// MSG_TIMEOUT: method has not been received fully (so call this function again until you get MSG_OK or MSG_ERROR)
	StreamMsg ReadRequest(unsigned int uTimeoutMs);

	// parses the request read by ReadRequest.
	// Returns true on success, false on failure.
	// This function must be called after ReadRequest returns MSG_OK and before any other function is called.
	bool ParseRequest();

	// returns for example "GET"
	string GetMethod() const;

	// returns for example "/whatever?query#fragment";
	// (this string will still be escaped!)
	string GetURI() const;

	// returns an unescaped string
	string GetPath() const;	// "/whatever" from "/whatever?query#fragment"
	string GetQuery() const;	// "query" from "/whatever?query#fragment"
	string GetFragment() const;	// "fragment" from "/whatever?query#fragment"

	// returns for example "HTTP/1.1"
	string GetHTTPVersion() const;

	// Gets client headers such as "Host"
	// Returns false if no such header was sent by the client.
	bool GetHeader(string &strValue, const string &strName);

	// Reads in POST data.
	// Returns MSG_END if no more POST data is available (or no POST data was ever available).
	StreamMsg Recv(void *buffer, size_t stBytesCanRead, size_t *pstBytesRead, unsigned int uTimeoutMs);

	// sets headers in preparation to call SendHeaders
	bool SetHeaders(const mpo_httpd_headers *pHeaders);

	// This function sends headers to the client after GetMethod has been called.
	// Returns MSG_OK if all headers have been sent and it's time to call Send.
	// Returns MSG_TIMEOUT if not all headers have been sent (keep calling to send them).
	StreamMsg SendHeaders(unsigned int uTimeoutMs);

	StreamMsg Send(const void *buffer, size_t bytes_can_send, size_t *bytes_sent, unsigned int timeout_ms);

	string GetLastErrorStr();

	static string Unescape(const string &strEscaped);

	// if 'strSrcURI' is "/a/b/c?d#e" then
	//  'strDstPath' is "/a/b/c", 'strDstQuery' is "d" and 'strDstFragment' is "e"
	static bool ParseURI(string &strDstPath, string &strDstQuery, string &strDstFragment, const string &strSrcURI);

private:
	typedef enum
	{
		RDY4_REQUEST,
		RDY4_PARSE,
		RDY4_RECV_POST,	// optional, only happens if method is POST
		RDY4_SET_HEADER,
		RDY4_SEND_HEADER,
		RDY4_SEND_BODY,
		RDY4_DISCONNECT	// if we client requested that this connection be closed
	} HttpdStage;

	// internal function
	StreamMsg ReadPartialRequest(unsigned int uTimeoutMs);

	mpo_httpd();
	~mpo_httpd();

	void DeleteInstance();

	// This is called from GetInstance
	bool Init(INonblockingStream *pStream);

	// helper function that streams in bytes and adds them to m_strCurBuf for parsing
	StreamMsg ReadIntoBuf(unsigned int uTimeoutMs);

	// TODO : move these into a general text parse class
	bool GetNextLine(string &strDst, string &strSrc);
	bool GetNextWord(string &strDst, string &strSrc);

	/////////////////////////////////////////////

	INonblockingStream *m_pStream;

	HttpdStage m_stage;

	// general purpose input buffer to deal with streaming
	string m_strCurBuf;

	// general purpose output buffer
	string m_strCurOutBuf;

	// the full request from the client
	string m_strRequest;

	// from the request line
	string m_strMethod;
	string m_strURI;
	string m_strURIPath, m_strURIQuery, m_strURIFragment;	// parsed and escaped from m_strURI
	string m_strHTTPVersion;

	// to store various client headers
	MapNameValue m_mapNameValuePairs;

	const mpo_httpd_headers *m_pHeaders;

	// so we know when we've sent the entire content length
	MPO_UINT64 m_u64BodyBytesSent;

	// so we know when we've read the entire POST body
	MPO_UINT64 m_u64PostBytesRead;
	MPO_UINT64 m_u64PostBytes;	// total # of bytes in POST body

	// last error message that occurred
	string m_strLastErrorStr;
};

class mpo_httpd_listener;

typedef boost::shared_ptr<mpo_httpd_listener> MpoHttpdListenerSPtr;

class CListenerComm
{
public:
	CListenerComm() :
	  pInstance(NULL),
		  m_bParentRequestedQuit(false),
		  m_bChildHasQuit(false)
	  {
	  }

	  // So listener thread can use all of the instance's private variables.
	  // Since mpo_httpd_listener doesn't have any public non-static methods, it is safe for the listener thread to do this without using mutex protection.
	  mpo_httpd_listener *pInstance;

	  bool m_bParentRequestedQuit;	// set by parent thread when quit is requested
	  bool m_bChildHasQuit;	// set by child thread when it has quit
};

class IHttpdThreadComm
{
public:
	// Returns true if parent thread has requested that we quit.  (current thread should shutdown ASAP if this is true)
	virtual bool IsQuitRequested() = 0;

	// returns pointer to httpd interface that child thread must use
	virtual Impo_httpd *GetHttpdInterface() = 0;

	// returns pointer to user-defined data, passed in via listener GetInstance call (can be NULL)
	virtual void *GetUserData() = 0;
};

class CHttpdThreadComm;

typedef boost::shared_ptr<CHttpdThreadComm> CHttpdThreadCommSPtr;

class CHttpdThreadComm : public IHttpdThreadComm, public MpoDeleter
{
	friend class mpo_httpd_listener;	// so that the httpd_listener thread helper can access private member vars and so child thread can't mess with them

public:
	// The reason I made this so it must be instantiated via GetInstance is to protect syncronization between child and parent for the member variables that this class contains.
	// Having it instantiated one time and passed around as a shared pointer does this.  Using the class's copy constructor does NOT.
	static CHttpdThreadCommSPtr GetInstance()
	{
		return CHttpdThreadCommSPtr(new CHttpdThreadComm(), CHttpdThreadComm::deleter());
	}

	// Returns true if parent thread has requested that we quit.  (current thread should shutdown ASAP if this is true)
	bool IsQuitRequested() { return m_bParentRequestedQuit; }

	Impo_httpd *GetHttpdInterface() { return pHTTPD; }

	void *GetUserData() { return data; }

private:
	CHttpdThreadComm() :
		m_pCallback(NULL),
		m_bChildHasQuit(false),
		m_bParentRequestedQuit(false)
	   {
	   }

	   void DeleteInstance() { delete this; }

	Impo_httpd *pHTTPD;	// pointer to httpd instance that child thread must use
	void *data;	// pointer to user-defined data, passed in via listener GetInstance call (can be NULL)

   	nonblocking_sharedptr stream;	// this needs to be here to keep the stream from disposing of itself (since the httpd instance doesn't store it)
	MpoHttpdSPtr HttpdSPtr;	// instance of Httpd class. private since it is a shared pointer, we don't want sync issues.
	void *(*m_pCallback)(void*);	// so that helper callback knows which callback to call
	bool m_bChildHasQuit;	// set by child thread when it has quit
	bool m_bParentRequestedQuit;	// set by parent thread when quit is requested
};

class EXPORT_ME mpo_httpd_listener : public MpoDeleter
{
public:
	static MpoHttpdListenerSPtr GetInstance(unsigned int uListenPort, void *(*pCallback)(void*), void *data = NULL);

	// returns the last error from a GetInstance call that failed
	static string GetLastGetInstanceError();

private:
	mpo_httpd_listener();
	~mpo_httpd_listener();

	void DeleteInstance();

	// This thread runs until the thread instance tells it to exit by setting a quit flag
	static void *ListenerThread(void *pListenerComm);

	static void *ListenerThreadHelper(void *pHttpThreadComm);

	///////////////////////////////////

	void *(*m_pCallback)(void *);	// callback called when new HTTPD connection is established
	void *data;	// user-defined data passed to callback

	// to accept connections
	mpo_server m_srv;

	// to communicate with child thread
	CListenerComm m_comm;

	// thread ID of listener thread
	mpo_threadID m_listenerID;

	// so caller knows why GetInstance failed
	static string m_strLastErr;
};

#endif // MPO_HTTPD_H
