#ifndef TEST_FRAMEWORK_H
#define TEST_FRAMEWORK_H

#include <string>
#include <list>
#include "mpo_types.h"

#include "mpo_dll.h"

#ifdef _WIN32
// get rid of warning C4251 which doesn't apply to us
#pragma warning( disable: 4251 )
#endif // WIN32

using namespace std;

typedef enum { MPO_EXCEPTION_REQUIRE_FAILED } mpo_exception_type;

#define TEST_CHECK(a) TestFrameWork::DoTest(#a, a, __LINE__, __FILE__)

#define TEST_REQUIRE(a) TestFrameWork::DoTest(#a, a, __LINE__, __FILE__); if (!a) throw MPO_EXCEPTION_REQUIRE_FAILED;

#define TEST_CHECK_EQUAL(a,b) TestFrameWork::DoTestEqual(a, b, __LINE__, __FILE__)

#define TEST_REQUIRE_EQUAL(a,b) TestFrameWork::DoTestEqual(a, b, __LINE__, __FILE__); if (a != b) throw MPO_EXCEPTION_REQUIRE_FAILED;

#define TEST_CHECK_NOT_EQUAL(a,b) TestFrameWork::DoTestNotEqual(a, b, __LINE__, __FILE__)

#define TEST_REQUIRE_NOT_EQUAL(a,b) TestFrameWork::DoTestNotEqual(a, b, __LINE__, __FILE__); if (a == b) throw MPO_EXCEPTION_REQUIRE_FAILED;

#define TEST_CHECK_THROW(a) { bool bCaught = false; try { a; } catch (...) { bCaught = true; } TestFrameWork::DoTest("expected exception from " #a, bCaught, __LINE__, __FILE__); }

struct entry_s
{
	string strDesc;	// description
	unsigned int uLine;	// line number
	string strFile;	// source file name
	string strTestCase;	// name of the test case
};

struct testcase_s
{
	string strName;	// name of test case
	void (*pTestPtr)();	// func pointer to test start
};

class EXPORT_ME TestFrameWork
{
public:

	// checks to make sure our testing variables are initialized and if they aren't, it initializes them
	static void CheckInit();

	static void AddTest(void (*pTestPtr)(), const string &strTestName, bool bExclusive);

	static void RunAllTests();

	static void DoUnhandledException(const string &strTestName);

	static int DoSummary();

	static void DoTest(const string &strDescription, const void *pResult, unsigned int uLine,
		const string &strSourceFile);

	static void DoTest(const string &strDescription, bool bResult, unsigned int uLine,
		const string &strSourceFile);

	// instantiators for the templates (add more here as needed)
	static void DoTestEqual(bool, bool, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(int, int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(unsigned int, int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(int, unsigned int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(int, long unsigned int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(long unsigned int, long unsigned int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(unsigned int, unsigned int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(long unsigned int, int, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(string, string, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(MPO_UINT64, MPO_UINT64, unsigned int uLine, const string &strSourceFile);
	static void DoTestEqual(MPO_UINT64, int, unsigned int uLine, const string &strSourceFile);

	static void DoTestNotEqual(int, int, unsigned int uLine, const string &strSourceFile);

private:

	template <class T1, class T2> static void DoTestEqualT(T1, T2, unsigned int uLine,
		const string &strSourceFile);

	template <class T1, class T2> static void DoTestNotEqualT(T1, T2, unsigned int uLine,
		const string &strSourceFile);

	static list <entry_s> m_lPassed;
	static list <entry_s> m_lFailed;

	// links to all test cases.
	// This is a pointer to work around the issue of constructors being called in an unpredictable order.
	static list <testcase_s>  *m_plTestCases;

	// if non-NULL, contains the name of an exclusive test to be run
	// (so that every single test doesn't have to be run when you're testing one specific thing)
	static string *m_pstrExclusiveTest;

};

EXPORT_ME void SetTestCaseName(const string &strTestCaseName);

// exclusive test case (no other tests will run)
#define TEST_CASE_EX(name) TEST_CASE_MAKER(name,true)

// normal test case (all tests will run)
#define TEST_CASE(name) TEST_CASE_MAKER(name,false)

// this will automatically run a test case
#define TEST_CASE_MAKER(name,exclusive) \
class name; \
class name	\
{	\
public:	\
	name() { TestFrameWork::AddTest(&run_test, #name, exclusive ); }	\
	static void run_test();	\
};	\
name name ; \
void name::run_test()

#endif // TEST_FRAMEWORK_H
